"""Top-level package for EpiStoch."""

__author__ = """Germán Riaño"""
__email__ = "griano@germanriano.com"
__version__ = "0.1.13"

from .seird_ph import *
from .sir_g import *
from .sir_phg import *
